
#include "entity.h"
#include "actor.h"


bool Entity::tryLoad( const char* name )
{
	// check if name does not contain illegal symbols
	const char* test = name;
	while( *test )
	{
		if( !(
			( *test >= 'a' && *test <= 'z' ) ||
			( *test >= 'A' && *test <= 'Z' ) ||
			( *test >= '0' && *test <= '9' ) ||
			*test == '_' || *test == '-'
			) )
		{
			sgs_Msg( C, SGS_WARNING, "entity name contains illegal symbols" );
			return false;
		}
		test++;
	}
	
	// try to load the script
	std::string fullname = "entities/";
	fullname += name;
	fullname += ".sgs";
	
	Handle( this ).push( C );
	sgs_StoreGlobal( C, "ENTITY" );
	if( SGS_FAILED( sgs_ExecFile( C, fullname.c_str() ) ) )
	{
		return false;
	}
	
	// for safety
	sgs_PushNull( C );
	sgs_StoreGlobal( C, "ENTITY" );
	
	m_type = name;
	return true;
}

Actor::Handle Entity::Spawn()
{
	printf( "Entity(%p)::Spawn\n", this );
	
	// should create the actor on some array
	// currently only returns the actor instead
	
	return Actor::Handle( SGS_PUSHCLASS( C, Actor, ( this ) ) );
}


int sgsfn_Entities_Create( SGS_CTX )
{
	char* name;
	
	SGSFN( "Entities::Create" );
	if( !sgs_LoadArgs( C, "s", &name ) )
		return 0;
	
	Entity* ent = SGS_PUSHCLASS( C, Entity, () );
	if( !ent->tryLoad( name ) )
	{
		// entity will be released as soon as the functions exits (it's referenced by the stack)
		return 0;
	}
	
	sgsHandle< Entity >( ent ).push( C );
	return 1;
}

static sgs_RegFuncConst entity_funcs[] =
{
	{ "Create", sgsfn_Entities_Create },
	SGS_RC_END(),
};

void sgs_register_entity( SGS_CTX )
{
	// store current env.
	sgs_Variable env;
	sgs_GetEnv( C, &env );
	
	// create "Entities" dict
	sgs_Variable entities;
	sgs_InitDict( C, &entities, 0 );
	// load it as current env.
	sgs_SetEnv( C, &entities );
	// add functions to it
	sgs_RegFuncConsts( C, entity_funcs, -1 );
	
	// reset env.
	sgs_SetEnv( C, &env );
	
	// add "Entities" to globals
	sgs_PushVariable( C, &entities );
	sgs_StoreGlobal( C, "Entities" );
	
	// release our handles to variables
	sgs_Release( C, &env );
	sgs_Release( C, &entities );
}
