
#ifndef SGS_INT_H_INCLUDED
#define SGS_INT_H_INCLUDED


#include "sgscript.h"
#include "sgs_util.h"


/*
	L I M I T S
	
	! some limits may cancel or reduce others, ..
	.. these are kept for the purposes of knowing the size of variables ..
	.. to use in various systems
	
	- number of constants in a function: 0 - 65535
	- number of instructions in a function: 0 - 65535
	- string size: 0 - 2^31-1
	- max. jump distance: 32767 instructions
	- max. loop depth: 65535 nested loops
		(no need to test for it since instruction count ..
		.. effectively limits this further)
	- bytecode size limit: 14 - 2^32-1
	- source size limit: 0 - 2^31-1
	- variable size limit: string size + overhead (<=256 bytes)
	- item type value limits: 8 bits
	
	- identifier size: 0 - 255
	- argument count: 0 - 255
	- returned value count: 0 - 255
	- total closure count in one function: 0 - 255
	- temporary variable count (incl. args): 0 - 255
	- useful line count limits in source file: 0 - 32767
	
	- stack size: 0 - 2^31-1
	- closure stack size: 0 - 2^31-1
	- array size: 0 - 2^31-1
	- hash table size: 0 - 2^31-1
*/


#ifdef __cplusplus
extern "C" {
#endif


#ifdef SGS_INTERNAL
#  define ST_NULL SGS_ST_NULL
#  define ST_RBRKL SGS_ST_RBRKL
#  define ST_RBRKR SGS_ST_RBRKR
#  define ST_SBRKL SGS_ST_SBRKL
#  define ST_SBRKR SGS_ST_SBRKR
#  define ST_CBRKL SGS_ST_CBRKL
#  define ST_CBRKR SGS_ST_CBRKR
#  define ST_ARGSEP SGS_ST_ARGSEP
#  define ST_STSEP SGS_ST_STSEP
#  define ST_PICKSEP SGS_ST_PICKSEP
#  define ST_IDENT SGS_ST_IDENT
#  define ST_KEYWORD SGS_ST_KEYWORD
#  define ST_NUMREAL SGS_ST_NUMREAL
#  define ST_NUMINT SGS_ST_NUMINT
#  define ST_STRING SGS_ST_STRING
#  define ST_OP_SEQ SGS_ST_OP_SEQ
#  define ST_OP_SNEQ SGS_ST_OP_SNEQ
#  define ST_OP_EQ SGS_ST_OP_EQ
#  define ST_OP_NEQ SGS_ST_OP_NEQ
#  define ST_OP_LEQ SGS_ST_OP_LEQ
#  define ST_OP_GEQ SGS_ST_OP_GEQ
#  define ST_OP_LESS SGS_ST_OP_LESS
#  define ST_OP_GRTR SGS_ST_OP_GRTR
#  define ST_OP_RWCMP SGS_ST_OP_RWCMP
#  define ST_OP_ADDEQ SGS_ST_OP_ADDEQ
#  define ST_OP_SUBEQ SGS_ST_OP_SUBEQ
#  define ST_OP_MULEQ SGS_ST_OP_MULEQ
#  define ST_OP_DIVEQ SGS_ST_OP_DIVEQ
#  define ST_OP_MODEQ SGS_ST_OP_MODEQ
#  define ST_OP_ANDEQ SGS_ST_OP_ANDEQ
#  define ST_OP_OREQ SGS_ST_OP_OREQ
#  define ST_OP_XOREQ SGS_ST_OP_XOREQ
#  define ST_OP_LSHEQ SGS_ST_OP_LSHEQ
#  define ST_OP_RSHEQ SGS_ST_OP_RSHEQ
#  define ST_OP_BLAEQ SGS_ST_OP_BLAEQ
#  define ST_OP_BLOEQ SGS_ST_OP_BLOEQ
#  define ST_OP_CATEQ SGS_ST_OP_CATEQ
#  define ST_OP_SET SGS_ST_OP_SET
#  define ST_OP_ERSUP SGS_ST_OP_ERSUP
#  define ST_OP_BLAND SGS_ST_OP_BLAND
#  define ST_OP_BLOR SGS_ST_OP_BLOR
#  define ST_OP_ADD SGS_ST_OP_ADD
#  define ST_OP_SUB SGS_ST_OP_SUB
#  define ST_OP_MUL SGS_ST_OP_MUL
#  define ST_OP_DIV SGS_ST_OP_DIV
#  define ST_OP_MOD SGS_ST_OP_MOD
#  define ST_OP_AND SGS_ST_OP_AND
#  define ST_OP_OR SGS_ST_OP_OR
#  define ST_OP_XOR SGS_ST_OP_XOR
#  define ST_OP_LSH SGS_ST_OP_LSH
#  define ST_OP_RSH SGS_ST_OP_RSH
#  define ST_OP_MMBR SGS_ST_OP_MMBR
#  define ST_OP_CAT SGS_ST_OP_CAT
#  define ST_OP_NOT SGS_ST_OP_NOT
#  define ST_OP_INV SGS_ST_OP_INV
#  define ST_OP_INC SGS_ST_OP_INC
#  define ST_OP_DEC SGS_ST_OP_DEC
#  define ST_ISOP SGS_ST_ISOP
#  define ST_OP_UNARY SGS_ST_OP_UNARY
#  define ST_OP_BINARY SGS_ST_OP_BINARY
#  define ST_OP_ASSIGN SGS_ST_OP_ASSIGN
#  define ST_OP_BINMUL SGS_ST_OP_BINMUL
#  define ST_OP_BINADD SGS_ST_OP_BINADD
#  define ST_OP_BINOPS SGS_ST_OP_BINOPS
#  define ST_OP_COMP SGS_ST_OP_COMP
#  define ST_OP_BOOL SGS_ST_OP_BOOL
#  define ST_ISSPEC SGS_ST_ISSPEC
#  define ST_READINT SGS_ST_READINT
#  define ST_READLN SGS_ST_READLN

/* #  define TokenType sgs_TokenType */
#  define TokenList sgs_TokenList

#  define SFT_NULL SGS_SFT_NULL
#  define SFT_IDENT SGS_SFT_IDENT
#  define SFT_KEYWORD SGS_SFT_KEYWORD
#  define SFT_CONST SGS_SFT_CONST
#  define SFT_OPER SGS_SFT_OPER
#  define SFT_OPER_P SGS_SFT_OPER_P
#  define SFT_FCALL SGS_SFT_FCALL
#  define SFT_INDEX SGS_SFT_INDEX
#  define SFT_MIDXSET SGS_SFT_MIDXSET
#  define SFT_MPROPSET SGS_SFT_MPROPSET
#  define SFT_ARGMT SGS_SFT_ARGMT
#  define SFT_ARGLIST SGS_SFT_ARGLIST
#  define SFT_VARLIST SGS_SFT_VARLIST
#  define SFT_GVLIST SGS_SFT_GVLIST
#  define SFT_USELIST SGS_SFT_USELIST
#  define SFT_EXPLIST SGS_SFT_EXPLIST
#  define SFT_ARRLIST SGS_SFT_ARRLIST
#  define SFT_MAPLIST SGS_SFT_MAPLIST
#  define SFT_RETURN SGS_SFT_RETURN
#  define SFT_BLOCK SGS_SFT_BLOCK
#  define SFT_IFELSE SGS_SFT_IFELSE
#  define SFT_WHILE SGS_SFT_WHILE
#  define SFT_DOWHILE SGS_SFT_DOWHILE
#  define SFT_FOR SGS_SFT_FOR
#  define SFT_FOREACH SGS_SFT_FOREACH
#  define SFT_BREAK SGS_SFT_BREAK
#  define SFT_CONT SGS_SFT_CONT
#  define SFT_FUNC SGS_SFT_FUNC
#  define FTNode sgs_FTNode

#  define CompFunc sgs_CompFunc
#  define rcpos_t sgs_rcpos_t

/* VM */
#  define CONSTVAR SGS_CONSTVAR
#  define CONSTENC SGS_CONSTENC
#  define CONSTDEC SGS_CONSTDEC

#  define SI_NOP SGS_SI_NOP
#  define SI_PUSH SGS_SI_PUSH
#  define SI_INT SGS_SI_INT
#  define SI_RETN SGS_SI_RETN
#  define SI_JUMP SGS_SI_JUMP
#  define SI_JMPT SGS_SI_JMPT
#  define SI_JMPF SGS_SI_JMPF
#  define SI_CALL SGS_SI_CALL
#  define SI_FORPREP SGS_SI_FORPREP
#  define SI_FORLOAD SGS_SI_FORLOAD
#  define SI_FORJUMP SGS_SI_FORJUMP
#  define SI_LOADCONST SGS_SI_LOADCONST
#  define SI_GETVAR SGS_SI_GETVAR
#  define SI_SETVAR SGS_SI_SETVAR
#  define SI_GETPROP SGS_SI_GETPROP
#  define SI_SETPROP SGS_SI_SETPROP
#  define SI_GETINDEX SGS_SI_GETINDEX
#  define SI_SETINDEX SGS_SI_SETINDEX
#  define SI_GENCLSR SGS_SI_GENCLSR
#  define SI_PUSHCLSR SGS_SI_PUSHCLSR
#  define SI_MAKECLSR SGS_SI_MAKECLSR
#  define SI_GETCLSR SGS_SI_GETCLSR
#  define SI_SETCLSR SGS_SI_SETCLSR
#  define SI_SET SGS_SI_SET
#  define SI_MCONCAT SGS_SI_MCONCAT
#  define SI_CONCAT SGS_SI_CONCAT
#  define SI_NEGATE SGS_SI_NEGATE
#  define SI_BOOL_INV SGS_SI_BOOL_INV
#  define SI_INVERT SGS_SI_INVERT
#  define SI_INC SGS_SI_INC
#  define SI_DEC SGS_SI_DEC
#  define SI_ADD SGS_SI_ADD
#  define SI_SUB SGS_SI_SUB
#  define SI_MUL SGS_SI_MUL
#  define SI_DIV SGS_SI_DIV
#  define SI_MOD SGS_SI_MOD
#  define SI_AND SGS_SI_AND
#  define SI_OR SGS_SI_OR
#  define SI_XOR SGS_SI_XOR
#  define SI_LSH SGS_SI_LSH
#  define SI_RSH SGS_SI_RSH
#  define SI_SEQ SGS_SI_SEQ
#  define SI_SNEQ SGS_SI_SNEQ
#  define SI_EQ SGS_SI_EQ
#  define SI_NEQ SGS_SI_NEQ
#  define SI_LT SGS_SI_LT
#  define SI_GTE SGS_SI_GTE
#  define SI_GT SGS_SI_GT
#  define SI_LTE SGS_SI_LTE
#  define SI_RAWCMP SGS_SI_RAWCMP
#  define SI_ARRAY SGS_SI_ARRAY
#  define SI_DICT SGS_SI_DICT

#  define instr_t sgs_instr_t

#  define INSTR_SIZE SGS_INSTR_SIZE
#  define INSTR_OFF_OP SGS_INSTR_OFF_OP
#  define INSTR_OFF_A SGS_INSTR_OFF_A
#  define INSTR_OFF_B SGS_INSTR_OFF_B
#  define INSTR_OFF_C SGS_INSTR_OFF_C
#  define INSTR_OFF_E SGS_INSTR_OFF_E
#  define INSTR_MASK_OP SGS_INSTR_MASK_OP
#  define INSTR_MASK_A SGS_INSTR_MASK_A
#  define INSTR_MASK_B SGS_INSTR_MASK_B
#  define INSTR_MASK_C SGS_INSTR_MASK_C
#  define INSTR_MASK_E SGS_INSTR_MASK_E
#  define INSTR_GET_OP SGS_INSTR_GET_OP
#  define INSTR_GET_A SGS_INSTR_GET_A
#  define INSTR_GET_B SGS_INSTR_GET_B
#  define INSTR_GET_C SGS_INSTR_GET_C
#  define INSTR_GET_E SGS_INSTR_GET_E
#  define INSTR_MAKE_OP SGS_INSTR_MAKE_OP
#  define INSTR_MAKE_A SGS_INSTR_MAKE_A
#  define INSTR_MAKE_B SGS_INSTR_MAKE_B
#  define INSTR_MAKE_C SGS_INSTR_MAKE_C
#  define INSTR_MAKE_E SGS_INSTR_MAKE_E
#  define INSTR_MAKE SGS_INSTR_MAKE
#  define INSTR_MAKE_EX SGS_INSTR_MAKE_EX
#  define INSTR_RECOMB_E SGS_INSTR_RECOMB_E

#  define func_t sgs_iFunc
#  define func_consts sgs_func_consts
#  define func_bytecode sgs_func_bytecode
#  define func_c_consts sgs_func_c_consts
#  define func_c_bytecode sgs_func_c_bytecode
#  define string_t sgs_iStr
#  define str_cstr sgs_str_cstr
#  define str_c_cstr sgs_str_c_cstr
#  define var_cstr sgs_var_cstr
#  define object_t sgs_object_t
#  define var_destroy_object sgsVM_VarDestroyObject

#  define STACKFRAMESIZE SGS_STACKFRAMESIZE
#endif



/*
	Token stream
	<1 byte: type> <additional data> <2 bytes: line number>
*/
/*     special          id */
#define SGS_ST_NULL	    '\0'
#define SGS_ST_RBRKL    '('
#define SGS_ST_RBRKR    ')'
#define SGS_ST_SBRKL    '['
#define SGS_ST_SBRKR    ']'
#define SGS_ST_CBRKL    '{'
#define SGS_ST_CBRKR    '}'
#define SGS_ST_ARGSEP   ','
#define SGS_ST_STSEP    ';'
#define SGS_ST_PICKSEP  ':'
/*     other            id    additional data */
#define SGS_ST_IDENT    'N' /* 1 byte (string size), N bytes (string), not null-terminated */
#define SGS_ST_KEYWORD  'K' /* same as IDENT */
#define SGS_ST_NUMREAL  'R' /* 8 bytes (double) */
#define SGS_ST_NUMINT   'I' /* 8 bytes (int64) */
#define SGS_ST_STRING   'S' /* 4 bytes (string size), N bytes (string), not null-terminated */
/*     operators        id    type  */
#define SGS_ST_OP_SEQ   200 /* ===  */
#define SGS_ST_OP_SNEQ  201 /* !==  */
#define SGS_ST_OP_EQ    202 /* ==   */
#define SGS_ST_OP_NEQ   203 /* !=   */
#define SGS_ST_OP_LEQ   204 /* <=   */
#define SGS_ST_OP_GEQ   205 /* >=   */
#define SGS_ST_OP_LESS  206 /* <    */
#define SGS_ST_OP_GRTR  207 /* >    */
#define SGS_ST_OP_RWCMP 208 /* <=>  */
#define SGS_ST_OP_ADDEQ 209 /* +=   */
#define SGS_ST_OP_SUBEQ 210 /* -=   */
#define SGS_ST_OP_MULEQ 211 /* *=   */
#define SGS_ST_OP_DIVEQ 212 /* /=   */
#define SGS_ST_OP_MODEQ 213 /* %=   */
#define SGS_ST_OP_ANDEQ 214 /* &=   */
#define SGS_ST_OP_OREQ  215 /* |=   */
#define SGS_ST_OP_XOREQ 216 /* ^=   */
#define SGS_ST_OP_LSHEQ 217 /* <<=  */
#define SGS_ST_OP_RSHEQ 218 /* >>=  */
#define SGS_ST_OP_BLAEQ 219 /* &&=  */
#define SGS_ST_OP_BLOEQ 220 /* ||=  */
#define SGS_ST_OP_CATEQ 221 /* $=   */
#define SGS_ST_OP_SET   222 /* =    */
#define SGS_ST_OP_ERSUP 223 /* @    */
#define SGS_ST_OP_BLAND 224 /* &&   */
#define SGS_ST_OP_BLOR  225 /* ||   */
#define SGS_ST_OP_ADD   226 /* +    */
#define SGS_ST_OP_SUB   227 /* -    */
#define SGS_ST_OP_MUL   228 /* *    */
#define SGS_ST_OP_DIV   229 /* /    */
#define SGS_ST_OP_MOD   230 /* %    */
#define SGS_ST_OP_AND   231 /* &    */
#define SGS_ST_OP_OR    232 /* |    */
#define SGS_ST_OP_XOR   233 /* ^    */
#define SGS_ST_OP_LSH   234 /* <<   */
#define SGS_ST_OP_RSH   235 /* >>   */
#define SGS_ST_OP_MMBR  236 /* .    */
#define SGS_ST_OP_CAT   237 /* $    */
#define SGS_ST_OP_NOT   238 /* !    */
#define SGS_ST_OP_INV   239 /* ~    */
#define SGS_ST_OP_INC   240 /* ++   */
#define SGS_ST_OP_DEC   241 /* --   */

#define SGS_ST_ISOP( chr )      ( (chr) >= 200 && (chr) <= 241 )
#define SGS_ST_OP_UNARY( chr )  ( (chr) == SGS_ST_OP_ERSUP || (chr) == SGS_ST_OP_ADD || \
	(chr) == SGS_ST_OP_SUB || ( (chr) >= SGS_ST_OP_NOT && (chr) <= SGS_ST_OP_DEC ) )
#define SGS_ST_OP_BINARY( chr ) ( (chr) >= 200 && (chr) <= 237 && (chr) != 223 )
#define SGS_ST_OP_ASSIGN( chr ) ( (chr) >= 209 && (chr) <= 222 )
#define SGS_ST_OP_BINMUL( chr ) ( (chr) == SGS_ST_OP_MUL || (chr) == SGS_ST_OP_DIV || (chr) == SGS_ST_OP_MOD )
#define SGS_ST_OP_BINADD( chr ) ( (chr) == SGS_ST_OP_ADD || (chr) == SGS_ST_OP_SUB )
#define SGS_ST_OP_BINOPS( chr ) ( (chr) >= SGS_ST_OP_AND && (chr) <= SGS_ST_OP_RSH )
#define SGS_ST_OP_COMP( chr )   ( (chr) >= 200 && (chr) <= 208 )
#define SGS_ST_OP_BOOL( chr )   ( (chr) == SGS_ST_OP_BLAEQ || (chr) == SGS_ST_OP_BLOEQ || \
	(chr) == SGS_ST_OP_BLAND || (chr) == SGS_ST_OP_BLOR )

#define SGS_ST_ISSPEC( chr )    isoneof( (chr), "()[]{},;:" )

#define SGS_ST_READINT( tgt, pos )   SGS_AS_INT32( tgt, pos )
#define SGS_ST_READLN( tgt, pos )    SGS_AS_( tgt, pos, sgs_LineNum )


typedef unsigned char sgs_TokenType;
typedef unsigned char* sgs_TokenList;

SGS_APIFUNC sgs_TokenList sgsT_Gen( SGS_CTX, const char* code, size_t length );
SGS_APIFUNC void          sgsT_Free( SGS_CTX, sgs_TokenList tlist );
SGS_APIFUNC sgs_TokenList sgsT_Next( sgs_TokenList tok );
SGS_APIFUNC sgs_LineNum   sgsT_LineNum( sgs_TokenList tok );

SGS_APIFUNC size_t sgsT_ListSize( sgs_TokenList tlist );
SGS_APIFUNC size_t sgsT_ListMemSize( sgs_TokenList tlist );

SGS_APIFUNC void sgsT_TokenString( SGS_CTX, sgs_MemBuf* out, sgs_TokenList tlist, sgs_TokenList tend, int xs );

SGS_APIFUNC void sgsT_DumpToken( sgs_TokenList tok );
SGS_APIFUNC void sgsT_DumpList( sgs_TokenList tlist, sgs_TokenList tend );


/*
	Function tree
*/

#define SGS_SFT_NULL    0
/* data */
#define SGS_SFT_IDENT   1
#define SGS_SFT_KEYWORD 2
#define SGS_SFT_CONST   3
/* expression parts */
#define SGS_SFT_OPER    4
#define SGS_SFT_OPER_P  5 /* post-op (inc/dec) */
#define SGS_SFT_FCALL   6
#define SGS_SFT_INDEX   7
#define SGS_SFT_MIDXSET 8
#define SGS_SFT_MPROPSET 9
/* statement data */
#define SGS_SFT_ARGMT   10
#define SGS_SFT_ARGLIST 11
#define SGS_SFT_VARLIST 12
#define SGS_SFT_GVLIST  13
#define SGS_SFT_USELIST 14
#define SGS_SFT_EXPLIST 15
#define SGS_SFT_ARRLIST 16
#define SGS_SFT_MAPLIST 17
#define SGS_SFT_RETURN  18
/* statement types */
#define SGS_SFT_BLOCK   21
#define SGS_SFT_IFELSE  22
#define SGS_SFT_WHILE   23
#define SGS_SFT_DOWHILE 24
#define SGS_SFT_FOR     25
#define SGS_SFT_FOREACH 26
#define SGS_SFT_BREAK   27
#define SGS_SFT_CONT    28
#define SGS_SFT_FUNC    30

typedef struct _sgs_FTNode sgs_FTNode;
struct _sgs_FTNode
{
	sgs_TokenList token;
	sgs_FTNode*   next;
	sgs_FTNode*   child;
	int           type;
};

SGS_APIFUNC void sgsFT_Destroy( SGS_CTX, sgs_FTNode* tree );

SGS_APIFUNC sgs_FTNode* sgsFT_Compile( SGS_CTX, sgs_TokenList tlist );
SGS_APIFUNC void sgsFT_Dump( sgs_FTNode* tree );



/*
	Intermediate function
	- keeps the function data between compiler stages
*/

typedef
struct _sgs_CompFunc
{
	sgs_MemBuf consts;
	sgs_MemBuf code;
	sgs_MemBuf lnbuf;
	uint8_t gotthis; /* bool */
	uint8_t numargs; /* guaranteed to be <= 255 by a test in `preparse_arglist` */
	uint8_t numtmp; /* reg. count (0-255, incl. numargs) - numargs (0-255) */
	uint8_t numclsr;
}
sgs_CompFunc;


/* - bytecode generator */
SGS_APIFUNC sgs_CompFunc* sgsBC_Generate( SGS_CTX, sgs_FTNode* tree );
SGS_APIFUNC void sgsBC_Dump( sgs_CompFunc* func );
SGS_APIFUNC void sgsBC_DumpEx( const char* constptr, size_t constsize,
	const char* codeptr, size_t codesize );
SGS_APIFUNC void sgsBC_Free( SGS_CTX, sgs_CompFunc* func );


/*
	Serialized bytecode
*/
SGS_APIFUNC int sgsBC_Func2Buf( SGS_CTX, sgs_CompFunc* func, sgs_MemBuf* outbuf );

/* assumes headers have already been validated (except size) but are still in the buffer */
SGS_APIFUNC const char* sgsBC_Buf2Func( SGS_CTX, const char* fn,
	const char* buf, size_t size, sgs_CompFunc** outfunc );

/* validates header size and bytes one by one (except last flag byte)
-- will return header_size on success and failed byte position on failure */
SGS_APIFUNC int sgsBC_ValidateHeader( const char* buf, size_t size );
#define SGS_HEADER_SIZE 14
#define SGS_MIN_BC_SIZE 14 + SGS_HEADER_SIZE
#define SGSBC_FLAG_LITTLE_ENDIAN 0x01


/*
	Virtual Machine
*/

#define SGS_CONSTVAR( x ) ((x)&0x100)
#define SGS_CONSTENC( x ) ((x)|0x100)
#define SGS_CONSTDEC( x ) ((x)&0xff)

#define SGS_INT_ERRSUP_INC 1
#define SGS_INT_ERRSUP_DEC 2

typedef enum sgs_Instruction_e
{
	SGS_SI_NOP = 0,

	SGS_SI_PUSH,     /* (B:src)                 push register/constant */
	SGS_SI_INT,      /* (C:type)                invoke a system state change */

	SGS_SI_RETN,     /* (A:N)                   exit current frame of execution, preserve N output arguments */
	SGS_SI_JUMP,     /* (E:off)                 add to instruction pointer */
	SGS_SI_JMPT,     /* (C:src, E:off)          jump (add to instr.ptr.) if true */
	SGS_SI_JMPF,     /* (C:src, E:off)          jump (add to instr.ptr.) if false */
	SGS_SI_CALL,     /* (A:exp, B:from, C:to)   call a variable */

	SGS_SI_FORPREP,  /* (A:out, B:src)          retrieves the iterator to work the object */
	SGS_SI_FORLOAD,  /* (A:iter, B:oky, C:ovl)  retrieves current key & value from iterator */
	SGS_SI_FORJUMP,  /* (C:iter, E:off)         advances and jumps if next key exists */

	SGS_SI_LOADCONST,/* (C:out, E:off)          load a constant to a register */
	SGS_SI_GETVAR,   /* (A:out, B:name)         <varname> => <value> */
	SGS_SI_SETVAR,   /* (B:name, C:src)         <varname> <value> => set <value> to <varname> */
	SGS_SI_GETPROP,  /* (A:out, B:var, C:name)  <var> <prop> => <var> */
	SGS_SI_SETPROP,  /* (A:var, B:name, C:src)  <var> <prop> <value> => set a <prop> of <var> to <value> */
	SGS_SI_GETINDEX, /* -- || -- */
	SGS_SI_SETINDEX, /* -- || -- */

	/* closures */
	SGS_SI_GENCLSR,  /* (A:N)                   generates closure variables for the stack frame */
	SGS_SI_PUSHCLSR, /* (A:src)                 pick closure variable for closure creation */
	SGS_SI_MAKECLSR, /* (A:out, B:from, C:N)    creates a closure object containing closure vars */
	SGS_SI_GETCLSR,  /* (A:out, B:which)        loads data from the specified closure */
	SGS_SI_SETCLSR,  /* (B:which, C:src)        stores data to the specified closure */

	/* operators */
	/*
		A: (A:out, B:s1, C:s2)
		B: (A:out, B:s1)
	*/
	SGS_SI_SET,      /* B */
	SGS_SI_MCONCAT,  /* (A:out, B:N) */
	SGS_SI_CONCAT,   /* A */
	SGS_SI_NEGATE,   /* B */
	SGS_SI_BOOL_INV,
	SGS_SI_INVERT,

	SGS_SI_INC,      /* B */
	SGS_SI_DEC,
	SGS_SI_ADD,      /* A */
	SGS_SI_SUB,
	SGS_SI_MUL,
	SGS_SI_DIV,
	SGS_SI_MOD,

	SGS_SI_AND,      /* A */
	SGS_SI_OR,
	SGS_SI_XOR,
	SGS_SI_LSH,
	SGS_SI_RSH,

	SGS_SI_SEQ,      /* A */
	SGS_SI_SNEQ,
	SGS_SI_EQ,
	SGS_SI_NEQ,
	SGS_SI_LT,
	SGS_SI_GTE,
	SGS_SI_GT,
	SGS_SI_LTE,
	SGS_SI_RAWCMP,

	/* specials */
	SGS_SI_ARRAY,    /* (A:out, B:args) */
	SGS_SI_DICT,     /* -- || -- */

	SGS_SI_COUNT
}
sgs_Instruction;


typedef uint32_t sgs_instr_t;

/*
	instruction data: 32 bits
	OOOOOOAAAAAAAABBBBBBBBBCCCCCCCCC
	OOOOOOEEEEEEEEEEEEEEEEECCCCCCCCC
*/

#define SGS_INSTR_SIZE    4

#define SGS_INSTR_OFF_OP  0
#define SGS_INSTR_OFF_A   6
#define SGS_INSTR_OFF_B   14
#define SGS_INSTR_OFF_C   23
#define SGS_INSTR_OFF_E   6

#define SGS_INSTR_MASK_OP 0x003f /* OP: 6 bits */
#define SGS_INSTR_MASK_A  0x00ff /* A:  8 bits */
#define SGS_INSTR_MASK_B  0x01ff /* B:  9 bits */
#define SGS_INSTR_MASK_C  0x01ff /* C:  9 bits */
#define SGS_INSTR_MASK_E  0x0001ffff /* E: 17 bits */

#define SGS_INSTR_GET_OP( x )  ((int) (((x) >> SGS_INSTR_OFF_OP) & SGS_INSTR_MASK_OP))
#define SGS_INSTR_GET_A( x )   ((int) (((x) >> SGS_INSTR_OFF_A ) & SGS_INSTR_MASK_A ))
#define SGS_INSTR_GET_B( x )   ((int) (((x) >> SGS_INSTR_OFF_B ) & SGS_INSTR_MASK_B ))
#define SGS_INSTR_GET_C( x )   ((int) (((x) >> SGS_INSTR_OFF_C ) & SGS_INSTR_MASK_C ))
#define SGS_INSTR_GET_E( x )   ((int) (((x) >> SGS_INSTR_OFF_E ) & SGS_INSTR_MASK_E ))

#define SGS_INSTR_MAKE_OP( x ) ((sgs_instr_t) (((x) & SGS_INSTR_MASK_OP) << SGS_INSTR_OFF_OP))
#define SGS_INSTR_MAKE_A( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_A ) << SGS_INSTR_OFF_A ))
#define SGS_INSTR_MAKE_B( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_B ) << SGS_INSTR_OFF_B ))
#define SGS_INSTR_MAKE_C( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_C ) << SGS_INSTR_OFF_C ))
#define SGS_INSTR_MAKE_E( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_E ) << SGS_INSTR_OFF_E ))

#define SGS_INSTR_MAKE( op, a, b, c ) \
	( SGS_INSTR_MAKE_OP( op ) | SGS_INSTR_MAKE_A( a ) | SGS_INSTR_MAKE_B( b ) | SGS_INSTR_MAKE_C( c ) )
#define SGS_INSTR_MAKE_EX( op, ex, c ) \
	( SGS_INSTR_MAKE_OP( op ) | SGS_INSTR_MAKE_E( ex ) | SGS_INSTR_MAKE_C( c ) )
#define SGS_INSTR_RECOMB_E( a, b ) ( ( ( b ) << 8 ) | ( a ) )


struct _sgs_iFunc
{
	int32_t refcount;
	uint32_t size;
	uint32_t instr_off;
	uint8_t gotthis;
	uint8_t numargs;
	uint8_t numtmp;
	uint8_t numclsr;
	sgs_LineNum linenum;
	sgs_LineNum* lineinfo;
	sgs_MemBuf funcname;
	sgs_MemBuf filename;
};

SGS_CASSERT( sizeof(sgs_Variable) % 4 == 0, variable_object_chaining_issue );
SGS_CASSERT( sizeof(sgs_iFunc) % 4 == 0, ifunc_object_chaining_issue );
SGS_CASSERT( sizeof(sgs_iStr) % 8 == 0, istr_object_storage_compat_issue );

#define sgs_func_consts( pfn )   ((sgs_Variable*)SGS_ASSUME_ALIGNED(((sgs_iFunc*)(pfn))+1,16))
#define sgs_func_bytecode( pfn ) ((sgs_instr_t*)(sgs_func_consts(pfn)+pfn->instr_off/sizeof(sgs_Variable)))
#define sgs_func_c_consts( pfn )   ((const sgs_Variable*)SGS_ASSUME_ALIGNED(((const sgs_iFunc*)(pfn))+1,16))
#define sgs_func_c_bytecode( pfn ) \
	((const sgs_instr_t*)(sgs_func_c_consts(pfn)+pfn->instr_off/sizeof(sgs_Variable)))

#define sgs_object_t sgs_VarObj

typedef struct _sgs_Closure sgs_Closure;
struct _sgs_Closure
{
	int32_t refcount;
	sgs_Variable var;
};


/* VM interface */
void sgsVM_VarCreateString( SGS_CTX, sgs_Variable* out, const char* str, sgs_SizeVal len );
void sgsVM_VarDestroyObject( SGS_CTX, sgs_object_t* O );

size_t sgsVM_VarSize( const sgs_Variable* var );
void sgsVM_VarDump( const sgs_Variable* var );

void sgsVM_StackDump( SGS_CTX );

int sgsVM_ExecFn( SGS_CTX, int numtmp, void* code, size_t codesize,
	void* data, size_t datasize, int clean, uint16_t* T );
int sgsVM_VarCall( SGS_CTX, sgs_Variable* var, int args, int clsr, int expect, int gotthis );
void sgsVM_PushClosures( SGS_CTX, sgs_Closure** cls, int num );


sgs_Variable* sgsVM_VarMake_Dict();

int sgsVM_RegStdLibs( SGS_CTX );

int sgs_specfn_call( SGS_CTX );
int sgs_specfn_apply( SGS_CTX );


/*
	Context handling
*/

typedef struct _sgs_BreakInfo sgs_BreakInfo;
struct _sgs_BreakInfo
{
	sgs_BreakInfo* next;
	uint32_t jdoff;  /* jump data offset */
	uint16_t numlp;  /* which loop */
	uint8_t  iscont; /* is a "continue"? */
};

/* register / constant position */
typedef int32_t sgs_rcpos_t;

typedef
struct _sgs_FuncCtx
{
	int32_t func;
	sgs_rcpos_t regs, lastreg;
	sgs_MemBuf vars;
	sgs_MemBuf gvars;
	sgs_MemBuf clsr;
	int inclsr, outclsr;
	int32_t loops;
	sgs_BreakInfo* binfo;
}
sgs_FuncCtx;

typedef
struct _sgs_ObjPoolItem
{
	sgs_VarObj* obj;
	uint32_t appsize;
}
sgs_ObjPoolItem;

typedef sgs_Variable* sgs_VarPtr;

struct _sgs_Context
{
	uint32_t      version;
	uint32_t      apiversion;

	/* output */
	sgs_OutputFunc output_fn; /* output function */
	void*         output_ctx; /* output context */
	sgs_OutputFunc erroutput_fn; /* error output function */
	void*         erroutput_ctx; /* error output context */

	/* info output */
	int32_t       minlev;
	int32_t       apilev;
	sgs_MsgFunc   msg_fn;  /* messaging function */
	void*         msg_ctx; /* messaging context */
	int           last_errno;

	/* hook */
	sgs_HookFunc  hook_fn;
	void*         hook_ctx;

	/* memory */
	sgs_MemFunc   memfunc;
	void*         mfuserdata;
	size_t        memsize;
	size_t        numallocs;
	size_t        numfrees;
	size_t        numblocks;

	/* compilation */
	uint32_t      state;
	sgs_FuncCtx*  fctx;      /* ByteCodeGen */
	const char*   filename;  /* filename of currently compiled code */
	sgs_VHTable   stringtable; /* string constant caching hash table */

	/* virtual machine */
	/* > main stack */
	sgs_VarPtr    stack_base;
	uint32_t      stack_mem;
	sgs_VarPtr    stack_off;
	sgs_VarPtr    stack_top;
	
	/* > closure pointer stack */
	sgs_Closure** clstk_base;
	sgs_Closure** clstk_off;
	sgs_Closure** clstk_top;
	uint32_t      clstk_mem;
	
	/* > stack frame info */
	sgs_StackFrame* sf_first;
	sgs_StackFrame* sf_last;
	sgs_StackFrame* sf_cached;
	int           sf_count;
	
	/* > _G (global variable dictionary) */
	sgs_VarObj*   _G;
	
	/* > object info */
	sgs_object_t* objs;
	int32_t       objcount;
	/* >> object GC */
	uint8_t       redblue;
	sgs_VarPtr    gclist;
	uint16_t      gclist_size;
	uint16_t      gcrun;
	/* >> object pool */
	sgs_ObjPoolItem* objpool_data;
	int32_t       objpool_size;
	
	/* type interface table */
	sgs_VHTable   typetable;
};

#define SGS_STACKFRAMESIZE ((sgs_StkIdx)(C->stack_top - C->stack_off))


#ifdef SGS_INTERNAL_STRINGTABLES

static const char* sgs_ErrNames[] =
{
	"SUCCESS", "ENOTFND", "ECOMP", "ENOTOBJ",
	"ENOTSUP", "EBOUNDS", "EINVAL", "EINPROC"
};

static const char* sgs_VarNames[] =
{
	"null",
	"bool",
	"int",
	"real",
	"string",
	"function",
	"C function",
	"object",
	"ptr",
};

static const char* sgs_OpNames[] =
{
	"nop",  "push", "int",
	"return", "jump", "jump_if_true", "jump_if_false", "call",
	"for_prep", "for_load", "for_jump", "loadconst", "getvar", "setvar",
	"getprop", "setprop", "getindex", "setindex",
	"genclsr", "pushclsr", "makeclsr", "getclsr", "setclsr",
	"set", "mconcat", "concat", "negate", "bool_inv", "invert",
	"inc", "dec", "add", "sub", "mul", "div", "mod",
	"and", "or", "xor", "lsh", "rsh",
	"seq", "sneq", "eq", "neq", "lt", "gte", "gt", "lte", "rawcmp",
	"array", "dict"
};

#endif


int sgsSTD_PostInit( SGS_CTX );
int sgsSTD_MakeArray( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt );
int sgsSTD_MakeDict( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt );
int sgsSTD_MakeMap( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt );
int sgsSTD_MakeClosure( SGS_CTX, sgs_Variable* func, uint32_t clc );
int sgsSTD_GlobalInit( SGS_CTX );
int sgsSTD_GlobalFree( SGS_CTX );
int sgsSTD_GlobalGet( SGS_CTX, sgs_Variable* out, sgs_Variable* idx );
int sgsSTD_GlobalSet( SGS_CTX, sgs_Variable* idx, sgs_Variable* val );
int sgsSTD_GlobalGC( SGS_CTX );
int sgsSTD_GlobalIter( SGS_CTX, sgs_VHTVar** outp, sgs_VHTVar** outpend );



/* STRUCTS */

typedef struct sgsstd_array_header_s
{
	sgs_SizeVal size;
	sgs_SizeVal mem;
	sgs_Variable* data;
}
sgsstd_array_header_t;

typedef struct sgsstd_array_iter_s
{
	sgs_Variable ref;
	sgs_SizeVal size;
	sgs_SizeVal off;
}
sgsstd_array_iter_t;

extern sgs_ObjInterface sgsstd_array_iface[1];
extern sgs_ObjInterface sgsstd_array_iter_iface[1];
extern sgs_ObjInterface sgsstd_dict_iface[1];
extern sgs_ObjInterface sgsstd_dict_iter_iface[1];
extern sgs_ObjInterface sgsstd_map_iface[1];

#define SGSIFACE_ARRAY SGS_IF_DLL( sgsstd_array_iface, sgs_FindType( C, "array" ) )
#define SGSIFACE_DICT SGS_IF_DLL( sgsstd_dict_iface, sgs_FindType( C, "dict" ) )
#define SGSIFACE_MAP SGS_IF_DLL( sgsstd_map_iface, sgs_FindType( C, "map" ) )




#ifdef __cplusplus
}
#endif


#endif /* SGS_INT_H_INCLUDED */
